(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Nonreachability\<close>

theory Nonreachability
imports
  "../Rewriting/Tcap_Impl"
  "../Tree_Automata/Exact_Tree_Automata_Completion_Impl"
begin

lemma rsteps_imp_rev_rsteps:
  assumes "(s, t) \<in> (rstep R)\<^sup>*"
  shows "(t, s) \<in> (rstep (R\<inverse>))\<^sup>*"
using assms using rtrancl_converse by force

lemma rstep_subst_overapproximation:
  assumes "\<forall>(l, r) \<in> R. \<exists>(l', r') \<in> R'. \<exists>\<sigma>. l = l' \<cdot> \<sigma> \<and> r = r' \<cdot> \<sigma>"
  shows "rstep R \<subseteq> rstep R'"
using assms by force

definition "is_instance_rule r r' \<longleftrightarrow>
  (case match_list Var [(fst r', fst r), (snd r', snd r)] of
    Some \<sigma> \<Rightarrow> True
  | None \<Rightarrow> False)"

lemma is_instance_rule [simp]:
  "is_instance_rule r r' \<longleftrightarrow> (\<exists>\<sigma>. fst r = fst r' \<cdot> \<sigma> \<and> snd r = snd r' \<cdot> \<sigma>)"
by (auto simp: is_instance_rule_def dest!: match_list_sound match_list_complete split: option.splits)
   metis

definition "check_subst_overapproximation R R' =
  check_allm (\<lambda>r. check_exm (\<lambda>r'.
    check (is_instance_rule r r') id
  ) R' (\<lambda>_. shows ''growing rule for '' \<circ> shows_rule r \<circ> shows '' is missing'' \<circ> shows_nl)) R
  <+? (\<lambda>e. shows R' \<circ> shows_nl \<circ>
        shows ''is not an overapproximation of'' \<circ> shows_nl \<circ> shows R \<circ> shows_nl \<circ> e)"

lemma check_subst_overapproximation [simp]:
  "isOK (check_subst_overapproximation R R') \<longleftrightarrow>
    (\<forall>(l, r) \<in> set R. \<exists>(l', r') \<in> set R'. \<exists>\<sigma>. l = l' \<cdot> \<sigma> \<and> r = r' \<cdot> \<sigma>)"
by (force simp: check_subst_overapproximation_def)

datatype ('f, 'v) nonreachability_proof =
  Nonreachable_Tcap
| Nonreachable_ETAC "('f \<times> nat) list" "'f" "'f" "('f, 'v) etac_ta"
| Nonreachable_Subst_Approx "('f, 'v) rules" "('f, 'v) nonreachability_proof"
| Nonreachable_Reverse "('f, 'v) nonreachability_proof"

fun check_nonreachable
where
  "check_nonreachable R s t Nonreachable_Tcap =
    check (\<not> Ground_Context.match (tcapI R s) t)
      (shows ''could not show nonreachability via tcap'')"
| "check_nonreachable R s t (Nonreachable_ETAC F a c A) =
    check_etac_nonreachable F a c A R s t"
| "check_nonreachable R s t (Nonreachable_Subst_Approx R' p) =
  check_subst_overapproximation R R' \<then> check_nonreachable R' s t p"
| "check_nonreachable R s t (Nonreachable_Reverse p) =
  check_nonreachable (map (\<lambda>(x, y). (y, x)) R) t s p"

lemma check_nonreachable [simp]:
  assumes "isOK (check_nonreachable R s t p)"
  shows "\<not> (\<exists>\<sigma> \<tau>. (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*)"
using assms
proof (induct p arbitrary: R s t)
  case Nonreachable_Tcap
  with match_tcap_sound [of s _ t _ "set R"] show ?case by auto
next
  case (Nonreachable_ETAC F a c A)
  with check_etac_nonreachable [of F a c A R s t] show ?case by auto
next
  case (Nonreachable_Subst_Approx R' p)
  then have "(\<forall>(l, r) \<in> set R. \<exists>(l', r') \<in> set R'. \<exists>\<sigma>. l = l' \<cdot> \<sigma> \<and> r = r' \<cdot> \<sigma>)"
    and "\<not> (\<exists>\<sigma> \<tau>. (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R'))\<^sup>*)" by auto
  with rstep_subst_overapproximation [OF this(1), THEN rtrancl_mono] show ?case by blast
next
  case *: (Nonreachable_Reverse p)
  have [simp]: "(\<lambda>(x, y). (y, x)) ` set R = (set R)\<inverse>" by auto
  have "\<not> (\<exists>\<sigma> \<tau>. (t \<cdot> \<sigma>, s \<cdot> \<tau>) \<in> (rstep ((set R)\<inverse>))\<^sup>*)" using * by force
  with rsteps_imp_rev_rsteps [of "s \<cdot> \<sigma>" "t \<cdot> \<tau>" "set R" for \<sigma> \<tau>]
    show ?case by auto
qed

datatype ('f, 'v) nonjoinability_proof =
  Nonjoinable_Tcap
| Nonjoinable_Ground_NF "('f, 'v) nonreachability_proof"

fun check_nonjoinable
where
  "check_nonjoinable R s t Nonjoinable_Tcap =
    check (\<not> Ground_Context.unifiable (tcapI R s) (tcapI R t))
      (shows ''could not show nonjoinability via tcap'')"
| "check_nonjoinable R s t (Nonjoinable_Ground_NF p) =
    (if is_NF_trs R s \<and> ground s then check_nonreachable R t s p
    else if is_NF_trs R t \<and> ground t then check_nonreachable R s t p
    else error (shows ''non NF''))"

lemma check_nonjoinable [simp]:
  assumes "isOK (check_nonjoinable R s t p)"
  shows "\<not> (\<exists>\<sigma>. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep (set R))\<^sup>\<down>)"
proof (cases p)
  case Nonjoinable_Tcap
  with assms show ?thesis using join_imp_unifiable_tcaps [of s _ t] by auto
next
  case (Nonjoinable_Ground_NF q)
  then show ?thesis
    using assms by (auto dest!: check_nonreachable split: if_splits)
                   (metis NF_join_imp_reach join_sym ground_subst_apply)+
qed

end
